/************************************************************************
 *									*
 *				N O T I C E				*
 *									*
 *			Copyright Abandoned, 1987, Fred Fish		*
 *									*
 *	This previously copyrighted work has been placed into the	*
 *	public domain by the author (Fred Fish) and may be freely used	*
 *	for any purpose, private or commercial.  I would appreciate	*
 *	it, as a courtesy, if this notice is left in all copies and	*
 *	derivative works.  Thank you, and enjoy...			*
 *									*
 *	The author makes no warranty of any kind with respect to this	*
 *	product and explicitly disclaims any implied warranties of	*
 *	merchantability or fitness for any particular purpose.		*
 *									*
 ************************************************************************
 */


/*
 *  FUNCTION
 *
 *	exp   double precision exponential
 *
 *  KEY WORDS
 *
 *	exp
 *	machine independent routines
 *	math libraries
 *
 *  DESCRIPTION
 *
 *	Returns double precision exponential of double precision
 *	floating point number.
 *
 *  USAGE
 *
 *	double exp (x)
 *	double x;
 *
 *  REFERENCES
 *
 *	Fortran IV plus user's guide, Digital Equipment Corp. pp B-3
 *
 *	Computer Approximations, J.F. Hart et al, John Wiley & Sons,
 *	1968, pp. 96-104.
 *
 *  RESTRICTIONS
 *
 *	Inputs greater than log(MAXDOUBLE) result in overflow.
 *	Inputs less than log(MINDOUBLE) result in underflow.
 *
 *	The maximum relative error for the approximating polynomial
 *	is 10**(-16.4).  However, this assumes exact arithmetic
 *	in the polynomial evaluation.  Additional rounding and
 *	truncation errors may occur as the argument is reduced
 *	to the range over which the polynomial approximation
 *	is valid, and as the polynomial is evaluated using
 *	finite precision arithmetic.
 *	
 *  PROGRAMMER
 *
 *	Fred Fish
 *
 *  INTERNALS
 *
 *	Computes exponential from:
 *
 *		exp(x)	=	2**y  *  2**z  *  2**w
 *
 *	Where:
 *
 *		y	=	int ( x * log2(e) )
 *
 *		v	=	16 * frac ( x * log2(e))
 *
 *		z	=	(1/16) * int (v)
 *
 *		w	=	(1/16) * frac (v)
 *
 *	Note that:
 *
 *		0 =< v < 16
 *
 *		z = {0, 1/16, 2/16, ...15/16}
 *
 *		0 =< w < 1/16
 *
 *	Then:
 *
 *		2**z is looked up in a table of 2**0, 2**1/16, ...
 *
 *		2**w is computed from an approximation:
 *
 *			2**w	=  (A + B) / (A - B)
 *
 *			A	=  C + (D * w * w)
 *
 *			B	=  w * (E + (F * w * w))
 *
 *			C	=  20.8137711965230361973
 *
 *			D	=  1.0
 *
 *			E	=  7.2135034108448192083
 *
 *			F	=  0.057761135831801928
 *
 *		Coefficients are from HART, table #1121, pg 206.
 *
 *		Effective multiplication by 2**y is done by a
 *		floating point scale with y as scale argument.
 *
 */
 
#include <stdio.h>
#include <pmluser.h>
#include "pml.h"
 
# define C  20.8137711965230361973	/* Polynomial approx coeff.	*/
# define D  1.0				/* Polynomial approx coeff.	*/
# define E  7.2135034108448192083	/* Polynomial approx coeff.	*/
# define F  0.057761135831801928	/* Polynomial approx coeff.	*/
 

/************************************************************************
 *									*
 *	This table is fixed in size and reasonably hardware		*
 *	independent.  The given constants were generated on a 		*
 *	DECSYSTEM 20 using double precision FORTRAN.			*
 *									*
 ************************************************************************
 */
 
static double fpof2tbl[] = {
    1.00000000000000000000,		/*    2 ** 0/16		*/
    1.04427378242741384020,		/*    2 ** 1/16		*/
    1.09050773266525765930,		/*    2 ** 2/16		*/
    1.13878863475669165390,		/*    2 ** 3/16		*/
    1.18920711500272106640,		/*    2 ** 4/16		*/
    1.24185781207348404890,		/*    2 ** 5/16		*/
    1.29683955465100966610,		/*    2 ** 6/16		*/
    1.35425554693689272850,		/*    2 ** 7/16		*/
    1.41421356237309504880,		/*    2 ** 8/16		*/
    1.47682614593949931110,		/*    2 ** 9/16		*/
    1.54221082540794082350,		/*    2 ** 10/16	*/
    1.61049033194925430820,		/*    2 ** 11/16	*/
    1.68179283050742908600,		/*    2 ** 12/16	*/
    1.75625216037329948340,		/*    2 ** 13/16	*/
    1.83400808640934246360,		/*    2 ** 14/16	*/
    1.91520656139714729380		/*    2 ** 15/16	*/
};

static char funcname[] = "exp";


double exp (x)
double x;
{
    register int y;
    register int index;
    auto double w;
    auto double v;
    auto double a;
    auto double b;
    auto double t;
    auto double temp;
    auto double wpof2;
    auto double zpof2;
    extern double dabs ();
    extern double ldexp ();
    extern double modf ();
    auto struct exception xcpt;
 
    DBUG_ENTER (funcname);
    DBUG_3 ("expin", "arg %le", x);
    if (x > LOGE_MAXDOUBLE) {
	xcpt.type = OVERFLOW;
	xcpt.name = funcname;
	xcpt.arg1 = x;
	if (!matherr (&xcpt)) {
	    fprintf (stderr, "%s: OVERFLOW error\n", funcname);
	    errno = ERANGE;
	    xcpt.retval = MAXDOUBLE;
	}
    } else if (x <= LOGE_MINDOUBLE) {
	xcpt.type = UNDERFLOW;
	xcpt.name = funcname;
	xcpt.arg1 = x;
	if (!matherr (&xcpt)) {
	    fprintf (stderr, "%s: OVERFLOW error\n", funcname);
	    errno = ERANGE;
	    xcpt.retval = 0.0;
	}
    } else {
	t = x * LOG2E;
	v = 16.0 * modf (t, &temp);
	y = temp;
	w = modf (v, &temp) / 16.0;
	index = temp;
	
	if (x < 0.0) {
	    zpof2 = 1.0 / fpof2tbl[-index];
	} else {
	    zpof2 = fpof2tbl[index];
	}
	a = C + (D * w * w);
	b = w * (E + (F * w * w));
	wpof2 = (a + b) / (a - b);
	xcpt.retval = ldexp ((wpof2 * zpof2), y);
    }
    DBUG_3 ("expout", "result %le", xcpt.retval);
    DBUG_RETURN (xcpt.retval);
}
