/************************************************************************
 *									*
 *				N O T I C E				*
 *									*
 *			Copyright Abandoned, 1987, Fred Fish		*
 *									*
 *	This previously copyrighted work has been placed into the	*
 *	public domain by the author (Fred Fish) and may be freely used	*
 *	for any purpose, private or commercial.  I would appreciate	*
 *	it, as a courtesy, if this notice is left in all copies and	*
 *	derivative works.  Thank you, and enjoy...			*
 *									*
 *	The author makes no warranty of any kind with respect to this	*
 *	product and explicitly disclaims any implied warranties of	*
 *	merchantability or fitness for any particular purpose.		*
 *									*
 ************************************************************************
 */


/*
 *  FUNCTION
 *
 *	log   double precision natural log
 *
 *  KEY WORDS
 *
 *	log
 *	machine independent routines
 *	math libraries
 *
 *  DESCRIPTION
 *
 *	Returns double precision natural log of double precision
 *	floating point argument.
 *
 *  USAGE
 *
 *	double log (x)
 *	double x;
 *
 *  REFERENCES
 *
 *	Computer Approximations, J.F. Hart et al, John Wiley & Sons,
 *	1968, pp. 105-111.
 *
 *  RESTRICTIONS
 *
 *	The absolute error in the approximating polynomial is
 *	10**(-19.38).  Note that contrary to DEC's assertion
 *	in the F4P user's guide, the error is absolute and not
 *	relative.
 *
 *	This error bound assumes exact arithmetic
 *	in the polynomial evaluation.  Additional rounding and
 *	truncation errors may occur as the argument is reduced
 *	to the range over which the polynomial approximation
 *	is valid, and as the polynomial is evaluated using
 *	finite-precision arithmetic.
 *	
 *  PROGRAMMER
 *
 *	Fred Fish
 *
 *  INTERNALS
 *
 *	Computes log(X) from:
 *
 *	  (1)	If argument is zero then flag an error
 *		and return minus infinity (or rather its
 *		machine representation).
 *
 *	  (2)	If argument is negative then flag an
 *		error and return minus infinity.
 *
 *	  (3)	Given that x = m * 2**k then extract
 *		mantissa m and exponent k.
 *
 *	  (4)	Transform m which is in range [0.5,1.0]
 *		to range [1/sqrt(2), sqrt(2)] by:
 *
 *			s = m * sqrt(2)
 *
 *	  (5)	Compute z = (s - 1) / (s + 1)
 *
 *	  (6)	Now use the approximation from HART
 *		page 111 to find log(s):
 *
 *		log(s) = z * ( P(z**2) / Q(z**2) )
 *
 *		Where:
 *
 *		P(z**2) =  SUM [ Pj * (z**(2*j)) ]
 *		over j = {0,1,2,3}
 *
 *		Q(z**2) =  SUM [ Qj * (z**(2*j)) ]
 *		over j = {0,1,2,3}
 *
 *		P0 =  -0.240139179559210509868484e2
 *		P1 =  0.30957292821537650062264e2
 *		P2 =  -0.96376909336868659324e1
 *		P3 =  0.4210873712179797145
 *		Q0 =  -0.120069589779605254717525e2
 *		Q1 =  0.19480966070088973051623e2
 *		Q2 =  -0.89111090279378312337e1
 *		Q3 =  1.0000
 *
 *		(coefficients from HART table #2705 pg 229)
 *
 *	  (7)	Finally, compute log(x) from:
 *
 *		log(x) = (k * log(2)) - log(sqrt(2)) + log(s)
 *
 */

#include <stdio.h>
#include <pmluser.h>
#include "pml.h"

static double log_pcoeffs[] = {
   -0.24013917955921050986e2,
    0.30957292821537650062e2,
   -0.96376909336868659324e1,
    0.4210873712179797145
};

static double log_qcoeffs[] = {
   -0.12006958977960525471e2,
    0.19480966070088973051e2,
   -0.89111090279378312337e1,
    1.0000
};

static char funcname[] = "log";


double log (x)
double x;
{
    auto int k;
    auto double s;
    auto double z;
    auto double zt2;
    auto double pqofz;
    auto struct exception xcpt;
    extern double frexp ();
    extern double poly ();

    DBUG_ENTER (funcname);
    DBUG_3 ("login", "arg %le", x);
    if (x == 0.0) {
	xcpt.type = SING;
	xcpt.name = funcname;
	xcpt.arg1 = x;
	if (!matherr (&xcpt)) {
	    fprintf (stderr, "%s: SINGULARITY error\n", funcname);
	    errno = EDOM;
	    xcpt.retval = -MAXDOUBLE;
	}
    } else if (x < 0.0) {
	xcpt.type = DOMAIN;
	xcpt.name = funcname;
	xcpt.arg1 = x;
	if (!matherr (&xcpt)) {
	    fprintf (stderr, "%s: DOMAIN error\n", funcname);
	    errno = EDOM;
	    xcpt.retval = -MAXDOUBLE;
	}
    } else {
	s = SQRT2 * frexp (x, &k);
	DBUG_3 ("log", "k = %d", k);
	DBUG_3 ("log", "s = %le", s);
	z = (s - 1.0) / (s + 1.0);
	DBUG_3 ("log", "z = %le", z);
	zt2 = z * z;
	DBUG_3 ("log", "zt2 = %le", zt2);
	pqofz = z * (poly (3, log_pcoeffs, zt2) / poly (3, log_qcoeffs, zt2));
	DBUG_3 ("pqofz", "pqofz = %le", pqofz);
	DBUG_3 ("log", "k = %d", k);
	DBUG_3 ("log", "LN2 = %le", LN2);
	x = k * LN2;
	DBUG_3 ("log", "x = %le", x);
	x -= LNSQRT2;
	DBUG_3 ("log", "x = %le", x);
	x += pqofz;
	DBUG_3 ("log", "x = %le", x);
	xcpt.retval = x;
    }
    DBUG_3 ("logout", "result %le", xcpt.retval);
    DBUG_RETURN (xcpt.retval);
}
